make.thrower <- function(expr) function() expr

catch<-function(tag, expr, wrap = FALSE) {
    "__CATCH_TAG__" <- as.character(tag)
    "__CATCH_THROWER__"<- make.thrower(return(get("__CATCH_VALUE__")))
    "__CATCH_VALUE__" <- wrap
    "__CATCH_WRAP__" <- wrap
    if (wrap)
        list(value = expr, throw = FALSE)
    else
        expr
}

##****parent.env with arg > 1 doesn't seem to work as advertized??
get.target.frame <- function(tag, name, not.found) {
    n <- sys.nframe()
    if (n > 1)
        for (i in (n-1):1) {
            env <- sys.frame(i)
            if (exists(name, env = env) &&
                get(name, env = env) == tag)
                return(env)
        }
    not.found(tag)
} 

dynamic.exists <- function(name) {
    n <- sys.nframe()
    if (n > 1)
        for (i in (n-1):1)
            if (exists(name, env = sys.frame(i)))
                return(TRUE)
    FALSE;
}


tag.not.found <- function(tag)
    stop(paste("no catch for tag \"", tag, "\"", sep = ""))

throw<-function(tag, value, no.tag = tag.not.found) {
    tag <- as.character(tag)
    env <- get.target.frame(tag, "__CATCH_TAG__", no.tag)
    if (get("__CATCH_WRAP__", env = env))
        value <- list(value = value, throw = TRUE)
    assign("__CATCH_VALUE__", value, env = env)
    fun <- get("__CATCH_THROWER__", env = env)
    fun()
}
print.exception <- function(e, ...)
    cat("<", class(e)[1], ": ", as.character(e), ">\n", sep="")

as.character.exception <- function(e, ...) {
    warning(paste("no as.character method for exception of class", class(e)))
    as.character(unclass(e))
}

simple.exception <- function(string) {
    class <- c("simple.exception", "exception")
    structure(list(string=as.character(string)), class=class)
}

as.character.simple.exception <- function(e, ...)
    e$string
reset.error.options <- function() {
    options(show.error.messages = TRUE) # ought to restore proper value
    options(error = NULL) # ought to restore proper value but can't seem to
}

set.error.options <- function() {
    options(show.error.messages = FALSE) # ought to save prefious value
    options(error = quote(error.converter())) # ought to save previous value but can't seem to
}

exists.try.catch <- function() dynamic.exists("__TRY_CATCH__")

default.exception.handler <- function(e)
    UseMethod("default.exception.handler", e)

default.exception.handler.exception <- function(e)
    stop(paste("unhandled exception (", class(e)[1],
               "): ", as.character(e), sep = ""))

raise.exception <- function(e) {
    reset.error.options()
    if (! inherits(e, "exception"))
        e <- simple.exception(as.character(e))
    if (exists.try.catch())
        throw("__TRY_CATCH__", e)
    else
        default.exception.handler(e)
}

try.catch <- function(expr, ..., finally = NULL) {
    set.error.options()
    on.exit({reset.error.options(); finally })
    result <- local({
        "__TRY_CATCH__" <- NULL
        catch("__TRY_CATCH__", expr, TRUE)
    })
    if (result$throw) {
        set.error.options() # turn back on because turned off in raise
        .Call("ResetErrorHandling") # is this the right place for this?
        handlers <- list(...)
        names <- names(handlers)
        e <- result$value
        for (i in seq(along = names))
            if (inherits(e, names[i]))
                return(handlers[[i]](e))
        on.exit()              # to get
        reset.error.options()  # things reasonable
        finally                # for stop calls going to default handler
        raise.exception(e)
    }
    else result$value
}
error.converter <- function() {
    reset.error.options() # turn off to avoid recursion
    raise.exception(simple.exception(geterrmessage()))
}
.First.lib <- function(lib, pkg)
    library.dynam(pkg, pkg, lib)
